from matplotlib import pyplot as plt
import pandas as pd
import seaborn as sns
import numpy as np
import math
from io import StringIO
import scipy.stats as stats
from scipy.stats import norm
import statsmodels.api as sm
import statsmodels.stats.api as sms
import statsmodels.stats.outliers_influence as sso
import statsmodels.formula.api as smf
from statsmodels.stats.stattools import durbin_watson as sdw
from . import table
from .model_building import *
from .model_building import _color_palette


_runs_test_table = pd.read_csv(table._raw_r_table, index_col=0)

_dwcrit_table = pd.read_csv(table._dwcrit_table, index_col=0)
_dwcrit_table25 = pd.read_csv(table._dwcrit_table25, index_col=0)


def durbin_watson(std_resid, n=None, k=None, option='positive', precision=4):
    """
    only detect first order autocorrelation between conseutive residuals in a time series

    + option = 'positive' / 'right-tail'
    + option = 'two-tail
    + option = 'negative' / 'left-tail'
    """
    rej_h_0_prompt = ""
    dwcrit_table = _dwcrit_table
    alpha = 0.05
    if option.lower()[0] == 'p' or option.lower()[0] == 'r':
        opt = 'r'
        rej_h_0_prompt = 'The data are positively first-order correlated'
    elif option.lower()[0] == 'n' or option.lower()[0] == 'l':
        opt = 'l'
        rej_h_0_prompt = 'The data are negatively first-order correlated'
    else:
        opt = 't'
        alpha = 0.025
        dwcrit_table = _dwcrit_table25
        rej_h_0_prompt = 'The data are first-order correlated'

    d = sdw(std_resid)
    d_l, d_u = dwcrit_table[(dwcrit_table['Sample Size'] == n) & (dwcrit_table['Number of terms (including the intercept)'] == k + 1)][['D  L',
                                                                                                                                        'D  U']].values.tolist()[0]

    d_result = f'''
d = {d:.{precision}f}
(d_l, d_u) = ({d_l:.{precision}f}, {d_u:.{precision}f})

with n = {n} and k = {k}
'''

    if opt == 'r':
        if d < d_l and d >= 0:
            flag = True
        elif d > d_u and d <= 4:
            flag = False
        elif d >= d_l and d <= d_u:
            flag = 'Inconclusive'
            rej_h_0_prompt = 'No meaning'
        else:
            print('Error occurred')
            print(d_result)
            return d, d_l, d_u

    elif opt == 'l':
        if d > 4 - d_l and d <= 4:
            flag = True
        elif d < 4 - d_u and d >= 2:
            flag = False
        elif d >= 4 - d_u and d <= 4 - d_l:
            flag = 'Inconclusive'
            rej_h_0_prompt = 'No meaning'
        else:
            print('Error occurred')
            print(d_result)
            return d, d_l, d_u
    elif opt == 't':
        if d < d_l or d > 4 - d_l:
            flag = True
        elif (d >= 4 - d_u and d <= 4 - d_l) or (d <= d_u and d >= d_l):
            flag = 'Inconclusive'
            rej_h_0_prompt = 'No meaning'
        elif d > d_u and d < 4 - d_u:
            flag = False
        else:
            print('Error occurred')
            print(d_result)
            return d, d_l, d_u

    results = f"""======== Durbin-Watson Test for First Order Correlation =======
d = {d:.{precision}f}
(d_l, d_u) = ({d_l:.{2}f}, {d_u:.{2}f})

with n = {n} and k = {k} / alpha = {alpha}

Reject H_0 ({rej_h_0_prompt}) → {flag}
"""
    print(results)
    return d, d_l, d_u


def add_margin(ax, x=0.05, y=0.05):

    xlim = ax.get_xlim()
    ylim = ax.get_ylim()

    xmargin = (xlim[1]-xlim[0])*x
    ymargin = (ylim[1]-ylim[0])*y

    ax.set_xlim(xlim[0]-xmargin, xlim[1]+xmargin)
    ax.set_ylim(ylim[0]-ymargin, ylim[1]+ymargin)


def inter_p_value(p_value):
    # interpretation
    if p_value >= 0 and p_value < 0.01:
        inter_p = 'Overwhelming Evidence'
    elif p_value >= 0.01 and p_value < 0.05:
        inter_p = 'Strong Evidence'
    elif p_value >= 0.05 and p_value < 0.1:
        inter_p = 'Weak Evidence'
    elif p_value >= .1:
        inter_p = 'No Evidence'
    return inter_p


def c_of_c_test(r, n, beta1=0, alpha=.05, precision=4, show=True, option='two-tail', **kwargs):
    """
    two-tail, right, left
    kwargs needs to contain b1 and std_err if used:
    >>> if beta1 != 0:
    >>>    t_value = (kwargs['b1'] - beta1) / kwargs['std_err']
    """
    opt = option.lower()[0]
    t_value = r * ((n-2)/(1 - r**2)) ** 0.5
    if beta1 != 0:
        t_value = (kwargs['b1'] - beta1) / kwargs['std_err']
    t_critical = stats.t.ppf(1 - alpha/2, n - 2)
    p_value = stats.t.sf(np.abs(t_value), n - 2)*2
    if opt == 't':
        # two-tail test
        option = 'Two-Tail Test'
        p_value = (1 - stats.t.cdf(t_value, df=n-2)) * 2
        if (p_value > 1):
            p_value = (stats.t.cdf(t_value, df=n-2)) * 2
        t_critical = stats.t.ppf(1 - alpha/2, df=n-2)
    else:
        if opt == 'l':
            option = 'One-Tail Test (left tail)'
            p_value = stats.t.cdf(t_value, df=n - 2)
            t_critical = stats.t.ppf(alpha, df=n - 2)
        elif opt == 'r':
            option = 'One-Tail Test (right tail)'
            p_value = stats.t.sf(t_value, df=n - 2)
            t_critical = stats.t.ppf(1 - alpha, df=n - 2)

    des = f"""r = {r:.{precision}f}
t (critical value) = {t_critical:.{precision}f}
t (observed value)  = {t_value:.{precision}f}
p-value = {p_value:.{precision}f} ({inter_p_value(p_value)})
    """
    if show:
        print(des)
    return t_value, t_critical, p_value, option


def multi_scatter_plot(row, col, df, x_names, y_name, figsize=(13, 7), hspace=0.2, wspace=0.2, scatter_show=True, corr_show=True):
    fig = plt.figure(figsize=figsize)
    fig.subplots_adjust(hspace=hspace, wspace=wspace)
    y_var = df[y_name]

    for i, name in enumerate(x_names):
        ax = fig.add_subplot(row, col, i + 1)
        # data = df[value_name][df[treatment_name] == name]
        x_var = df[name].values
        slope, intercept, r_value, p_value, std_err = stats.linregress(
            x_var, y_var)  # order matters

        ax = sns.regplot(x=name, y=y_name, data=df, ci=None, scatter_kws={'color': 'dodgerblue'}, line_kws={
                         'color': '#ffaa77', 'label': f"$\hat y = {intercept:.4f} + {slope:.4f} x$"})

        plt.legend()

        plt.title(f'Scatter Plot for {name} vs {y_name}')
        plt.xlabel(name)
        plt.ylabel(y_name)

        add_margin(ax, x=0.02, y=0.00)  # Call this after tsplot

    fig.tight_layout()
    if scatter_show:
        plt.show()
    else:
        plt.close()

    df_corr = df[[y_name] + x_names]
    corr = df_corr.corr()
    fig, ax = plt.subplots()
    _ = sns.heatmap(corr, annot=True)
    if corr_show:
        plt.show()
    else:
        plt.close()


def multicollinearity(df, x_names, y_name):
    df_corr = df[[y_name] + x_names]
    corr = df_corr.corr()
    coef = corr.iloc[:, 0].values

    def _highlight(val):
        color = 'salmon' if (abs(val) >= 0.7 and val !=
                             1 and val not in coef) else 'default'
        return 'background-color: %s' % color

    style_df = corr.style.applymap(_highlight)
    return style_df


def SimpleLinearRegressionPrediction(x_name=None, y_name=None, df=None, x=None, alpha=0.05, plot=True, kwargs={'title': 'Two Types of Intervals', 'xlabel': 'Independent Variable', 'ylabel': 'Dependent Variable'}):
    """

    """
    x1 = x
    if type(x_name) == str:
        x, y = df[x_name], df[y_name]
    else:
        x, y = x_name, y_name
    x_new = np.array([1, x1])
    X2 = sm.add_constant(x)
    olsmod = sm.OLS(y, X2)
    result_reg = olsmod.fit()
    y_head = np.dot(result_reg.params, x_new)
    (t_minus, t_plus) = stats.t.interval(
        alpha=(1.0 - alpha), df=result_reg.df_resid)
    cov_mat1 = np.cov(y, x)
    x_bar = x.mean()
    core1 = (1 / result_reg.nobs +
             (x1 - x_bar) ** 2 / (result_reg.nobs - 1) / cov_mat1[1, 1]) ** 0.5
    core2 = (1 + 1 / result_reg.nobs +
             (x1 - x_bar) ** 2 / (result_reg.nobs - 1) / cov_mat1[1, 1]) ** 0.5
    lower_bound = y_head + t_minus * (result_reg.mse_resid ** 0.5) * core1
    upper_bound = y_head + t_plus * (result_reg.mse_resid ** 0.5) * core1
    half_interval = t_plus * (result_reg.mse_resid ** 0.5) * core1
    lower_bound2 = y_head + t_minus * (result_reg.mse_resid ** 0.5) * core2
    upper_bound2 = y_head + t_plus * (result_reg.mse_resid ** 0.5) * core2
    half_interval2 = t_plus * (result_reg.mse_resid ** 0.5) * core2
    des = f"""======= Making Prediction =======
Make CI and PI predictions at mean of x = {x1}
y_head = {y_head}

Confidence interval for mean: [{lower_bound:.4f}, {upper_bound:.4f}]
    or {y_head:.4f}  ± {half_interval:.4f}
Prediction interval (confidence interval) for Exact Value: [{lower_bound2:.4f}, {upper_bound2:.4f}]
    or {y_head:.4f}  ± {half_interval2:.4f}
        """

    print(des)
    CI_PI = {'CI': [lower_bound, upper_bound],
             'PI': [lower_bound2, upper_bound2]}
    try:
        df_sorted = df.sort_values([x_name])
        df_res = smf.ols(f'{y_name}~ {x_name}', data=df_sorted).fit()
        x = df_sorted[x_name].values
        y = df_sorted[y_name].values
        fig, ax = plt.subplots()
        st, data, ss3 = sso.summary_table(df_res, alpha=alpha)
        fittedvalues = data[:, 2]
        predict_mean_se = data[:, 3]
        predict_mean_ci_low, predict_mean_ci_upp = data[:, 4:6].T
        predict_ci_low, predict_ci_upp = data[:, 6:8].T
        plt.plot(x, y, 'o', color='gray')
        plt.plot(x, fittedvalues, 'g-', lw=0.5, label='Model')
        plt.plot(x, predict_mean_ci_low, 'r-',
                 lw=0.4, label='Confidence Interval')
        plt.plot(x, predict_mean_ci_upp, 'r-', lw=0.4)
        plt.plot(x, predict_ci_low, 'b--', lw=0.4,
                 label='Prediction Interval')
        plt.plot(x, predict_ci_upp, 'b--', lw=0.4)
        circle_rad = 12
        ax.plot((x1), (y_head), 'o', color='gold',
                label=f'$(x, \hat y)$ = ({x1}, {y_head:.4f})')
        ax.plot((x1), (y_head), 'o', ms=circle_rad *
                1.2, mec='gold', mfc='none', mew=2)
        plt.title(kwargs['title'])
        plt.xlabel(kwargs['xlabel'])
        plt.ylabel(kwargs['ylabel'])
        plt.legend()
        ax.legend(loc='center left', bbox_to_anchor=(
            1, 0.5), fancybox=True, shadow=True)
        CI_PI['fig'] = fig
        CI_PI['ax'] = ax
        if plot:
            plt.show()
        else:
            plt.close(fig)
    except:
        print('Indepent variable (x) and Dependent variable (y) should be passed in as "strings". Additional DataFrame is also required.')
    return CI_PI


def SimpleLinearRegressionOutlier(x_name=None, y_name=None, df=None, outlier=True, influential=True, plot=True, display_df=True, **kwargs):
    """
    kwargs:

    + Pass in xlabel to specify the x label for individual residual plot
    + pass in s_xlabel to specify the x label for the scatter plot
    + pass in s_ylabel to specify the y label for the scatter plot
    """

    df_res = smf.ols(f'{y_name}~ {x_name}', data=df).fit()
    simple_table, data, ss3 = sso.summary_table(df_res, alpha=0.05)
    std_resid = data[:, 10]
    return_dict = {}
    if outlier:
        df_w_std_resid = df.copy().reset_index().rename(
            columns={'index': 'ID'})
        df_w_std_resid['ID'] += 1
        df_w_std_resid['Std. Resid'] = std_resid
        filter = (df_w_std_resid['Std. Resid'] < -
                  2) | (df_w_std_resid['Std. Resid'] > 2)
        df_w_std_resid = df_w_std_resid[filter]
        # display(df_w_std_resid)
        return_dict['df_w_std_resid'] = df_w_std_resid
        return_dict['outlier_filter'] = filter

        if display_df:
            print('Outliers:')
            display(df_w_std_resid)

        y_pre = data[:, 2]  # x

        fig, ax = plt.subplots()

        plt.plot(y_pre, std_resid, 'o', color='gray')
        plt.axhline(y=0, color='blue')
        plt.axhline(y=2, color='red')
        plt.axhline(y=-2, color='red')

        circle_rad = 12
        for i in df_w_std_resid.index:
            x_id = i
            y_id = data[x_id, 2]
            sr_id = data[x_id, 10]

            ax.plot((y_id), (sr_id), 'o',
                    ms=circle_rad * 1.5, mec='violet', mfc='none', mew=2)
        ax.plot([], [], 'o', ms=circle_rad * 0.5, mec='violet',
                mfc='none', mew=2, label='Outliers')

        plt.title('Standardized Residual Plot - Outliers in Violet Circle')
        try:
            xlabel = kwargs['xlabel']
        except KeyError:
            xlabel = f'Predicted {y_name}'
        plt.xlabel(xlabel)
        plt.ylabel('Standardized Residual')

        return_dict['outlier-fig'] = fig
        return_dict['outlier-ax'] = ax

        if plot:
            plt.show()
        else:
            plt.close()

    if influential:
        df_w_h = df.copy().reset_index().rename(columns={'index': 'ID'})
        df_w_h['ID'] += 1
        x_data = df[x_name].values
        y_data = df[y_name].values
        cov_mat1 = np.cov(y_data, x_data)
        x_data_bar = x_data.mean()
        nobs = len(x_data)
        h_val = 1 / nobs + (x_data - x_data_bar) ** 2 / \
            (nobs - 1) / cov_mat1[1, 1]
        df_w_h['h (leverage)'] = h_val
        filter = (df_w_h['h (leverage)'] > 6 / nobs)
        df_w_h = df_w_h[filter]
        return_dict['df_w_h'] = df_w_h
        return_dict['h cv'] = 6 / nobs
        return_dict['inf_filter'] = filter

        if display_df:
            print('Influential Observations:')
            display(df_w_h)

        y_pre = data[:, 2]  # x

        fig, ax = plt.subplots()

        plt.plot(y_pre, std_resid, 'o', color='gray')
        plt.axhline(y=0, color='blue')
        plt.axhline(y=2, color='red')
        plt.axhline(y=-2, color='red')

        # cir = plt.Circle((y_id, sr_id), 0.2, color='y',fill=False)
        circle_rad = 12

        for i in df_w_h.index:
            x_id = i
            y_id = data[x_id, 2]
            sr_id = data[x_id, 10]
        #     x_id, y_id, sr_id

            ax.plot((y_id), (sr_id), 'o',
                    ms=circle_rad * 1.5, mec='springgreen', mfc='none', mew=2)
        # ax.set_aspect('equal', adjustable='datalim')
        # ax.add_patch(cir)
        ax.plot([], [], 'o',
                ms=circle_rad * 0.5, mec='springgreen', mfc='none', mew=2, label='Influential Observations')

        plt.title(
            f'Standardized Residual Plot - Influential Observations ($h_i > {6 / nobs}$) in Spring Green')
        try:
            xlabel = kwargs['xlabel']
        except KeyError:
            xlabel = f'Predicted {y_name}'
        plt.xlabel(xlabel)
        plt.ylabel('Standardized Residual')
        # plt.legend()
        return_dict['inf-fig'] = fig
        return_dict['inf-ax'] = ax
        plt.show()

        if plot:
            plt.show()
        else:
            plt.close()

    # Scatter Plot with Two Variables and Fixed Margins (Seaborn included)
    fig, ax = plt.subplots()

    slope, intercept, r_value, p_value, std_err = stats.linregress(
        df[x_name], df[y_name])  # order matters

    ax = sns.regplot(x=x_name, y=y_name, data=df, ci=None, scatter_kws={'color': 'dodgerblue'}, line_kws={
                     'color': '#ffaa77', 'label': f"y = {intercept:.4f} + {slope:.4f} x"})

    circle_rad = 12

    try:
        for i in df_w_std_resid.index:
            x_id = i
            y_id = data[x_id, 1]
            x_value = df[x_name][x_id]

            ax.plot((x_value), (y_id), 'o',
                    ms=circle_rad * 1.5, mec='violet', mfc='none', mew=2)
        ax.plot([], [], 'o', ms=circle_rad * 0.5, mec='violet',
                mfc='none', mew=2, label='Outliers')
    except:
        pass
    try:
        for i in df_w_h.index:
            x_id = i
            y_id = data[x_id, 1]
            x_value = df[x_name][x_id]

            ax.plot((x_value), (y_id), 'o',
                    ms=circle_rad * 1.5, mec='springgreen', mfc='none', mew=2)
        ax.plot([], [], 'o',
                ms=circle_rad * 0.5, mec='springgreen', mfc='none', mew=2, label='Influential Observations')

    except:
        pass

    plt.legend()

    ax.legend(loc='center left', bbox_to_anchor=(
        1, 0.5), fancybox=True, shadow=True)

    plt.title('Scatter Plot')
    try:
        xlabel = kwargs['s_xlabel']
    except KeyError:
        xlabel = f'{x_name}'
    try:
        ylabel = kwargs['s_ylabel']
    except KeyError:
        ylabel = f'{y_name}'
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

    add_margin(ax, x=0.02, y=0.00)  # Call this after tsplot

    return_dict['all-fig'] = fig
    return_dict['all-ax'] = ax

    if plot:
        plt.show()
    else:
        plt.close()

    return return_dict


def SimpleLinearRegression(x_name=None, y_name=None, df=None, alpha=0.05, precision=4, show_summary=True, plot=False, assessment=True, slope_option='two-tail', beta1=0, coeff_option='two-tail', kwargs={'x': 0.02, 'y': 0.00, 'title': 'Scatter Plot'}):
    """
    df_result = smf.ols(f'{y_name}~ {x_name}', data=df).fit()
    """
    slope, intercept, r_value, p_value, std_err_sb1 = stats.linregress(
        df[x_name], df[y_name])
    # flag = p_value < alpha
    result_dict = {'slope': slope, 'intercept': intercept,
                   'r_value': r_value, 'p_value': p_value}
    fig, ax = plt.subplots()

    ax = sns.regplot(x=x_name, y=y_name, data=df, ci=None, scatter_kws={'color': 'dodgerblue'}, line_kws={
        'color': '#ffaa77', 'label': f"y = {intercept:.4f} + {slope:.4f} x"})

    add_margin(ax, x=kwargs['x'], y=kwargs['y'])  # Call this after tsplot
    result_dict['fig'] = fig
    result_dict['ax'] = ax
    if plot:
        plt.legend()
        plt.title(kwargs['title'])
        if 'xlabel' in kwargs:
            plt.xlabel(kwargs['xlabel'])
        else:
            plt.xlabel(x_name)
        if 'ylabel' in kwargs:
            plt.ylabel(kwargs['ylabel'])
        else:
            plt.ylabel(y_name)
        plt.show()
    else:
        plt.close(fig)

    df_result = smf.ols(f'{y_name}~ {x_name}', data=df).fit()
    result_dict['result'] = df_result
    s_e = df_result.mse_resid ** 0.5
    ssr_value = df_result.ssr
    r_square = r_value ** 2
    s_t_value, s_t_critical, s_p_value, s_option = c_of_c_test(
        r_value, df_result.nobs, beta1=beta1, alpha=alpha, precision=precision, show=False, option=slope_option, kwargs={'b1': slope, 'std_err': std_err_sb1})
    s_flag = s_p_value < alpha
    t_t_value, t_t_critical, t_p_value, t_option = c_of_c_test(
        r_value, df_result.nobs, alpha=alpha, precision=precision, show=False, option=coeff_option)
    if t_option == 'Two-Tail Test':
        t_p_value = p_value
    flag = t_p_value < alpha
    ci_b1 = df_result.conf_int(alpha)[1:].values[0]
    result_dict['ci_b1'] = ci_b1

    if show_summary:
        print(df_result.summary())
        print()
    results = f"""======= Simple Linear Regression Results =======
Dep. Variable: {y_name}
No. of Observations: {int(df_result.nobs)}
Mean of Dep. Variable: {np.mean(df[y_name]):.{precision}f}
Standard Error: {s_e:.{precision}f}
SSR: {ssr_value:.{precision}f}
R-square: {r_square:.{precision}f}

Estimated model: y = {intercept:.{precision}f} + {slope:.{precision}f} x"""

    if assessment == True:
        results += f"""

**** t-Test of Slope <{s_option}> ****
t (observed value): {s_t_value:.{precision}f}
t (critical value): {s_t_critical:.{precision}f}
b1 (slope): {slope:.{precision}f}
p-value: {s_p_value:.{precision}f} ({inter_p_value(s_p_value)})

Reject H_0 (Has some kind of relationship between two variables) → {s_flag}

{(1-alpha) * 100}% confidence interval = [{ci_b1[0]:.4f}, {ci_b1[1]:.4f}]

**** t-Test of Correlation Coefficient <{t_option}> ****
t (observed value): {t_t_value:.{precision}f}
t (critical value): {t_t_critical:.{precision}f}
r: {r_value:.{precision}f}
p-value: {t_p_value:.{precision}f} ({inter_p_value(t_p_value)})
Reject H_0 (Has Linear Relationship) → {flag}"""
    print(results)
    result_dict['description'] = results

    return result_dict, results


def MultipleRegressionPrediction(x_names=None, y_name=None, df=None, x1=None, alpha=0.05, precision=4, ts=False, **kwargs):
    """
    if ts = True: x1 is required to be a list (iterable)
    """
    # print("make CI and PI prediction at mean of x = ", x1)
    if ts:
        tdf = df.copy()
        df_result = kwargs['df_result']
        time_name = kwargs['time_name']
        _, data, _ = sso.summary_table(df_result, alpha=0.05)
        new_t = np.array(x1)
        if len(x1) == 0:
            tdf[f'Pre_{y_name}'] = data[:, 2]
            return tdf
        elif len(x1) == 1:
            search_table = pd.DataFrame(x1, columns=x_names)
            x1 = [1] + x1[0]
            new_t = np.array([x1])
            # print(new_t)
        else:
            search_table = pd.DataFrame(x1, columns=x_names)
            new_t = sm.add_constant(new_t)

        pred_y = df_result.predict(new_t)
        tdf[f'Pre_{y_name}'] = data[:, 2]
        # tdf[x_name] = np.append(tdf[x_name], new_t)
        for i in range(search_table.shape[0]):
            append_dict = {}
            for x in x_names:
                print(x)
                append_dict[x] = search_table.iloc[i, :][x]
            append_dict[f'Pre_{y_name}'] = pred_y[i]
            tdf = tdf.append(
                append_dict, ignore_index=True)

        return tdf

    else:
        x1.insert(0, 1)
        X_data_T = np.array(df[x_names])
        X_data2 = sm.add_constant(X_data_T)
        olsmod = sm.OLS(df[y_name], X_data2)
        result_reg = olsmod.fit()
        # print(result_reg.params, x1)
        y_head = np.dot(result_reg.params, x1)
        # print("y_head = ", y_head)
        (t_minus, t_plus) = stats.t.interval(
            alpha=(1.0 - alpha), df=result_reg.df_resid)
        core1 = (result_reg.mse_resid * np.matmul(x1,
                                                  np.linalg.solve(np.matmul(X_data2.T, X_data2), x1))) ** 0.5
        lower_bound = y_head + t_minus * core1
        upper_bound = y_head + t_plus * core1
        # print("confidence interval of mean = [%0.4f, %0.4f] " % (
        #     lower_bound, upper_bound))
        core2 = (result_reg.mse_resid * (1 + np.matmul(x1,
                                                       np.linalg.solve(np.matmul(X_data2.T, X_data2), x1)))) ** 0.5
        lower_bound2 = y_head + t_minus * core2
        upper_bound2 = y_head + t_plus * core2

        result = f"""======= Making Prediction =======
    Make CI and PI predictions at mean of x = {x1}
    y_head = {y_head}

    ----------------
    {pd.DataFrame(result_reg.params, columns=['coef'])}
    ----------------

    Confidence interval for mean: [{lower_bound:.{precision}f}, {upper_bound:.{precision}f}]
    Prediction interval (confidence interval) for Exact Value: [{lower_bound2:.{precision}f}, {upper_bound2:.{precision}f}]
    """
        print(result)
        CI_PI = {'CI': [lower_bound, upper_bound],
                 'PI': [lower_bound2, upper_bound2]}
        return CI_PI


def MultipleRegressionOutlier(x_names=None, y_name=None, df=None, std_resid=None, outlier=True, influential=True, cook=True, alpha=.05, plot=True, display_df=True, **kwargs):
    """
    Pass in x_names, y_name, and df as well

    kwargs:

    + Pass in xlabel to specify the x label for individual residual plot
    + pass in s_xlabel to specify the x label for the scatter plot
    + pass in s_ylabel to specify the y label for the scatter plot
    """

    # df_res = smf.ols(f'{y_name}~ {x_name}', data=df).fit()
    y_data = df[y_name]
    X_data_T = np.array(df[x_names])
    X_data = pd.DataFrame(X_data_T, columns=x_names)
    X_data_update = sm.add_constant(X_data)
    olsmod = sm.OLS(y_data, X_data_update)
    reg_result = olsmod.fit()
    if std_resid is not None:
        pass
    else:
        reg_result = olsmod.fit()
        simple_table, data, ss3 = sso.summary_table(reg_result, alpha=alpha)
        std_resid = data[:, 10]

    return_dict = {}
    if outlier:
        df_w_std_resid = df.copy().reset_index().rename(
            columns={'index': 'ID'})
        df_w_std_resid['ID'] += 1
        df_w_std_resid['Std. Resid'] = std_resid
        filter = (df_w_std_resid['Std. Resid'] < -
                  2) | (df_w_std_resid['Std. Resid'] > 2)
        df_w_std_resid = df_w_std_resid[filter]
        # display(df_w_std_resid)
        return_dict['df_w_std_resid'] = df_w_std_resid
        return_dict['outlier_filter'] = filter

        if display_df:
            print('Outliers:')
            display(df_w_std_resid)

        y_pre = data[:, 2]  # x

        fig, ax = plt.subplots()

        plt.plot(y_pre, std_resid, 'o', color='gray')
        plt.axhline(y=0, color='blue')
        plt.axhline(y=2, color='red')
        plt.axhline(y=-2, color='red')

        circle_rad = 12
        for i in df_w_std_resid.index:
            x_id = i
            y_id = data[x_id, 2]
            sr_id = data[x_id, 10]

            ax.plot((y_id), (sr_id), 'o',
                    ms=circle_rad * 1.5, mec='violet', mfc='none', mew=2)
        ax.plot([], [], 'o', ms=circle_rad * 0.5, mec='violet',
                mfc='none', mew=2, label='Outliers')

        plt.title('Standardized Residual Plot - Outliers in Violet Circle')
        try:
            xlabel = kwargs['xlabel']
        except KeyError:
            xlabel = f'Predicted {y_name}'
        plt.xlabel(xlabel)
        plt.ylabel('Standardized Residual')

        return_dict['outlier-fig'] = fig
        return_dict['outlier-ax'] = ax

        if plot:
            plt.show()
        else:
            plt.close()

    if influential:
        x_data = df[x_names].values
        y_data = df[y_name].values
        x_data = sm.add_constant(x_data)
        H = np.matmul(x_data, np.linalg.solve(
            np.matmul(x_data.T, x_data), x_data.T))
        df_w_h = df.copy().reset_index().rename(columns={'index': 'ID'})
        df_w_h['ID'] += 1
        df_w_h['h_ii'] = np.diagonal(H)
        k = len(x_names)
        n = len(df_w_h['h_ii'])
        h_level = 3 * (k+1) / n
        # print("h_level = ", h_level)
        # print(" \n")
        filter = (df_w_h['h_ii'] > h_level)
        # print("Influential Observations by hi = \n")
        # print(df_w_h['h_ii'].loc[filter])
        df_w_h = df_w_h[filter]
        return_dict['df_w_h'] = df_w_h
        return_dict['h cv'] = h_level
        return_dict['inf_filter'] = filter

        if display_df:
            print('Influential Observations:')
            display(df_w_h)

        y_pre = data[:, 2]  # x

        fig, ax = plt.subplots()

        plt.plot(y_pre, std_resid, 'o', color='gray')
        plt.axhline(y=0, color='blue')
        plt.axhline(y=2, color='red')
        plt.axhline(y=-2, color='red')

        # cir = plt.Circle((y_id, sr_id), 0.2, color='y',fill=False)
        circle_rad = 12

        for i in df_w_h.index:
            x_id = i
            y_id = data[x_id, 2]
            sr_id = data[x_id, 10]
        #     x_id, y_id, sr_id

            ax.plot((y_id), (sr_id), 'o',
                    ms=circle_rad * 1.5, mec='springgreen', mfc='none', mew=2)
        # ax.set_aspect('equal', adjustable='datalim')
        # ax.add_patch(cir)
        ax.plot([], [], 'o',
                ms=circle_rad * 0.5, mec='springgreen', mfc='none', mew=2, label='Influential Observations')
        ii = "ii"
        plt.title(
            f'Standardized Residual Plot - Influential Observations ($h_{ii} > {h_level}$) in Spring Green')
        try:
            xlabel = kwargs['xlabel']
        except KeyError:
            xlabel = f'Predicted {y_name}'
        plt.xlabel(xlabel)
        plt.ylabel('Standardized Residual')
        # plt.legend()
        return_dict['inf-fig'] = fig
        return_dict['inf-ax'] = ax
        plt.show()

        if plot:
            plt.show()
        else:
            plt.close()

    if cook:
        s2_e = reg_result.mse_resid
        k = reg_result.df_model
        y_a = data[:, 1]
        y_f = data[:, 2]
        x_data = df[x_names].values
        y_data = df[y_name].values
        x_data = sm.add_constant(x_data)
        H = np.matmul(x_data, np.linalg.solve(
            np.matmul(x_data.T, x_data), x_data.T))
        df_w_h = df.copy().reset_index().rename(columns={'index': 'ID'})
        df_w_h['ID'] += 1
        df_w_h['h_ii'] = np.diagonal(H)
        h_i = df_w_h['h_ii']

        CD_arr = np.square(y_a - y_f) / s2_e / (k - 1) * \
            h_i / np.square(1 - h_i)

        CD = np.array(CD_arr)
        df_w_cd = df.copy().reset_index().rename(columns={'index': 'ID'})
        df_w_cd['ID'] += 1
        df_w_cd['CD'] = CD
        # display(df_w_cd)
        filter = (df_w_cd['CD'] > 1)
        # print("Influential Observations by Cook's Distances = \n")
        df_w_cd = df_w_cd[filter]
        return_dict['df_w_cd'] = df_w_cd
        # return_dict['h cv'] = h_level
        return_dict['cd_filter'] = filter

        if display_df:
            print('Influential Observations by Cook\'s Distance:')
            display(df_w_cd)

        y_pre = data[:, 2]  # x

        fig, ax = plt.subplots()

        plt.plot(y_pre, std_resid, 'o', color='gray')
        plt.axhline(y=0, color='blue')
        plt.axhline(y=2, color='red')
        plt.axhline(y=-2, color='red')

        # cir = plt.Circle((y_id, sr_id), 0.2, color='y',fill=False)
        circle_rad = 12

        for i in df_w_cd.index:
            x_id = i
            y_id = data[x_id, 2]
            sr_id = data[x_id, 10]
        #     x_id, y_id, sr_id

            ax.plot((y_id), (sr_id), 'o',
                    ms=circle_rad * 1.5, mec='goldenrod', mfc='none', mew=2)
        # ax.set_aspect('equal', adjustable='datalim')
        # ax.add_patch(cir)
        ax.plot([], [], 'o',
                ms=circle_rad * 0.5, mec='goldenrod', mfc='none', mew=2, label='Influential Observations by Cook\'s Distance')

        plt.title(
            f'Standardized Residual Plot - \nInfluential Observations by Cook\'s Distance ($D_i > 1$) in Golden Rod')
        try:
            xlabel = kwargs['xlabel']
        except KeyError:
            xlabel = f'Predicted {y_name}'
        plt.xlabel(xlabel)
        plt.ylabel('Standardized Residual')
        # plt.legend()
        return_dict['cd-fig'] = fig
        return_dict['cd-ax'] = ax
        plt.show()

        if plot:
            plt.show()
        else:
            plt.close()

    # # Scatter Plot with Two Variables and Fixed Margins (Seaborn included)
    # fig, ax = plt.subplots()

    # # slope, intercept, r_value, p_value, std_err = stats.linregress(
    # #     df[x_name], df[y_name])  # order matters

    # # ax = sns.scatterplot(x=x_name, y=y_name, data=df, ci=None, scatter_kws={'color': 'dodgerblue'}, line_kws={
    # #                  'color': '#ffaa77', 'label': f"y = {intercept:.4f} + {slope:.4f} x"})

    # circle_rad = 12

    # try:
    #     for i in df_w_std_resid.index:
    #         x_id = i
    #         y_id = data[x_id, 1]
    #         x_value = df[x_name][x_id]

    #         ax.plot((x_value), (y_id), 'o',
    #                 ms=circle_rad * 1.5, mec='violet', mfc='none', mew=2)
    #     ax.plot([], [], 'o', ms=circle_rad * 0.5, mec='violet',
    #             mfc='none', mew=2, label='Outliers')
    # except:
    #     pass
    # try:
    #     for i in df_w_h.index:
    #         x_id = i
    #         y_id = data[x_id, 1]
    #         x_value = df[x_name][x_id]

    #         ax.plot((x_value), (y_id), 'o',
    #                 ms=circle_rad * 1.5, mec='springgreen', mfc='none', mew=2)
    #     ax.plot([], [], 'o',
    #             ms=circle_rad * 0.5, mec='springgreen', mfc='none', mew=2, label='Influential Observations')

    # except:
    #     pass

    # plt.legend()

    # ax.legend(loc='center left', bbox_to_anchor=(
    #     1, 0.5), fancybox=True, shadow=True)

    # plt.title('Scatter Plot')
    # try:
    #     xlabel = kwargs['s_xlabel']
    # except KeyError:
    #     xlabel = f'{x_name}'
    # try:
    #     ylabel = kwargs['s_ylabel']
    # except KeyError:
    #     ylabel = f'{y_name}'
    # plt.xlabel(xlabel)
    # plt.ylabel(ylabel)

    # add_margin(ax, x=0.02, y=0.00)  # Call this after tsplot

    # return_dict['all-fig'] = fig
    # return_dict['all-ax'] = ax

    # if plot:
    #     plt.show()
    # else:
    #     plt.close()

    return return_dict


def MultipleRegressionResidual(x_names=None, y_name=None, df=None, reg_result=None, durbin_watson_test=True, durbin_watson_test_option='two-tail', alpha=0.05, precision=4):
    """
    np.mean(std_resid), np.std(std_resid, ddof=1)
    """
    if reg_result is not None:
        st, data, ss2 = sso.summary_table(reg_result, alpha=alpha)
    else:
        y_data = df[y_name]
        X_data_T = np.array(df[x_names])
        X_data = pd.DataFrame(X_data_T, columns=x_names)
        X_data_update = sm.add_constant(X_data)
        olsmod = sm.OLS(y_data, X_data_update)
        reg_result = olsmod.fit()
    st, data, ss2 = sso.summary_table(reg_result, alpha=alpha)
    y_pre = data[:, 2]
    y_id = data[:, 0]
    std_resid = data[:, 10]
    # std_resid
    stat, p_value = stats.shapiro(std_resid)
    # print('Statistics=%.4f, p=%.4f' % (stat, p))
    result = f'''======= Multiple Regression: Residual Analysis =======
☑ Normality Check: Using Shapiro-Wilk Test

Statistics = {stat:.{precision}f}, p-value = {p_value:.{precision}f} ({inter_p_value(p_value)})
Reject H_0 (Not normally distributed) → {p_value < alpha}

☑ Heteroscedasticity and Homoscedasticity
'''
    print(result)

    fig = plt.figure(figsize=(9, 4))
    fig.subplots_adjust(hspace=0.02, wspace=0.02)

    ax = fig.add_subplot(1, 2, 1)
    # data = df[value_name][df[treatment_name] == name]
    plt.plot(y_pre, std_resid, 'o', color='gray')
    plt.axhline(y=2, color='red', lw=0.8)
    plt.axhline(y=0, color='blue')
    plt.axhline(y=-2, color='red', lw=0.8)
    plt.title('Standardized Residual Plot')
    plt.xlabel('Predicted y value')
    plt.ylabel('Standardized Residual')

    ax = fig.add_subplot(1, 2, 2)
    plt.hist(std_resid, bins='auto')
    plt.title('Histogram of Standardized Residual')
    plt.xlabel('Standardized Residual')
    plt.ylabel('Frequency')
    add_margin(ax, x=0.02, y=0.00)  # Call this after tsplot

    fig.tight_layout()
    plt.show()

    print('\n☑ Non-x_name of the Error Variable Check\n')

    fig, ax = plt.subplots()
    plt.plot(y_id, std_resid, 'o', color='gray')
    plt.axhline(y=2, color='red', lw=0.8)
    plt.axhline(y=0, color='blue')
    plt.axhline(y=-2, color='red', lw=0.8)
    plt.title('Standardized Residual Plot')
    plt.xlabel('Observation No.')
    plt.ylabel('Standardized Residual')
    plt.show()

    runs_test(std_resid, cutoff='median', alpha=alpha, precision=precision)

    if durbin_watson_test:
        durbin_watson(std_resid, n=len(
            std_resid), k=reg_result.df_model, option=durbin_watson_test_option)

    return np.mean(std_resid), np.std(std_resid, ddof=1)


def MultipleRegression(x_names=None, y_name=None, df=None, alpha=0.05, precision=4, show_summary=True, assessment=True, t_test_c=0, t_test_option='two-tail'):
    """
    assessment is equivalent to t_test: boolean
    """
    res_dict = dict()
    y_data = df[y_name]
    X_data_T = np.array(df[x_names])
    X_data = pd.DataFrame(X_data_T, columns=x_names)
    X_data_update = sm.add_constant(X_data)
    olsmod = sm.OLS(y_data, X_data_update)
    df_result = olsmod.fit()
    ssr_value = df_result.ssr
    s_e = df_result.mse_resid ** 0.5
    res_dict['df_result'] = df_result

    st, data, ss2 = sso.summary_table(df_result, alpha=alpha)
    y_pre = data[:, 2]
    y_id = data[:, 0]
    std_resid = data[:, 10]

    res_dict['y_pre'] = y_pre
    res_dict['y_id'] = y_id
    res_dict['std_resid'] = std_resid
    if show_summary:
        print(df_result.summary())
        print()

    # get params = df_result.params
    # np.array(param)
    # coef_table = pd.DataFrame(df_result.summary().tables[1], dtype=str)
    # coef_table.columns = coef_table.iloc[0]
    # coef_table = coef_table.drop(index=0).reset_index(drop=True)
    # coef_col = coef_table['coef']
    coef_col = df_result.params

    # estimated model
    estimated_mod_str = f'ŷ = {coef_col[0]:.{precision}f}'

    for i, c in enumerate(coef_col[1:]):
        estimated_mod_str += f' + {c:.{precision}f} x{i + 1}'

    results = f"""======= Multiple Regression Results =======
Dep. Variable: {y_name}
No. of Observations (n): {int(df_result.nobs)}
No. of Ind. Vairable (k): {int(df_result.df_model)}
Mean of Dep. Variable: {np.mean(df[y_name]):.{precision}f}
Standard Deviation of Dep. Variable: {np.std(df[y_name], ddof=1):.{precision}f}
Standard Error: {s_e:.{precision}f} (ȳ = {np.mean(y_data):.{precision}f})
SSR: {ssr_value:.{precision}f}

R-square: {df_result.rsquared:.{precision}f}
Adjusted R-square: {df_result.rsquared_adj:.{precision}f}
Difference (≤ 0.06 {abs(df_result.rsquared - df_result.rsquared_adj) <= 0.06}): {df_result.rsquared - df_result.rsquared_adj}

Estimated model: {estimated_mod_str}"""
    results += f'''

<F-test>
F(observed value): {df_result.fvalue: .{precision}f}
p-value: {df_result.f_pvalue: .{precision}f} ({inter_p_value(df_result.f_pvalue)})
Reject H_0 (The model is valid: at least one beta_i ≠ 0) → {df_result.f_pvalue < alpha}
'''
    if assessment:
        # alpha =.05
        # t_test_c = 0
        try:
            i = iter(t_test_c)
        except:
            t_test_c = [t_test_c] * (int(df_result.df_model) + 1)
        conf_int_df = df_result.conf_int(alpha=alpha)
        conf_int_df.columns = [f'[{alpha/2}', f'{1 - alpha/2}]']
        t_test_df = pd.DataFrame(df_result.params, columns=['coef (βi)'])
        t_test_df['p-value'] = df_result.pvalues
        t_test_df = t_test_df.merge(
            conf_int_df, left_index=True, right_index=True)
        t_test_df['H_0'] = [
            f'β{i} = {t_test_c[i]}' for i in range(int(df_result.df_model) + 1)]
        (t_test_c > conf_int_df.iloc[:, 0])
        # option = 'two-tail'
        if t_test_option[0].lower() == 't':
            t_test_df['Reject H_0'] = (t_test_c < conf_int_df.iloc[:, 0]) | (
                t_test_c > conf_int_df.iloc[:, 1])
        elif t_test_option[0].lower() == 'l':
            t_test_df['Reject H_0'] = (t_test_c < conf_int_df.iloc[:, 0]) & (
                t_test_c < conf_int_df.iloc[:, 1])
        elif t_test_option[0].lower() == 'r':
            t_test_df['Reject H_0'] = (t_test_c > conf_int_df.iloc[:, 0]) & (
                t_test_c > conf_int_df.iloc[:, 1])
        # t_test_df

        res_dict['t_test_df'] = t_test_df

        results += f'''
<t-test for each βi>
{t_test_df.round(precision)}
'''
    print(results)
    return res_dict, results


def runs_test(x, cutoff='median', alpha=0.05, precision=4, show_table=False):
    if cutoff == 'median':
        cutoff = np.median(x)
    elif cutoff == 'mean':
        cutoff = np.mean(x)

    indicator = (x >= cutoff)
    n1 = np.sum(x >= cutoff)
    n2 = np.sum(x < cutoff)
    n_runs = 1

    for i in range(1, len(indicator)):
        if indicator[i] != indicator[i - 1]:
            n_runs += 1

    if n1 <= 20 and n2 <= 20:
        if show_table:
            display(_runs_test_table)
        if n1 > n2:
            big_n = n1
            small_n = n2
        else:
            big_n = n2
            small_n = n1

        res = (_runs_test_table.loc[small_n, str(big_n)])
        lb, ub = tuple(map(int, res.strip('()').split(', ')))
        if n_runs < ub and n_runs > lb:
            flag = False
        else:
            flag = True
        des = f"""======= Runs Test =======
(Both n1 ({n1}) and n2 ({n2}) <= 20) → Check Runs Test Table
Runs = {n_runs}
Lower bound and Upper bound = [{lb}, {ub}]
Reject H_0 (Randomness does not exist) → {flag}
"""
        print(des)
        return flag, ub, lb

    else:
        mu_r = 2 * n1 * n2 / (n1 + n2) + 1
        sigma_r = math.sqrt(2 * n1 * n2 * (2 * n1 * n2 -
                                           n1 - n2) / ((n1 + n2) ** 2 * (n1 + n2 - 1)))
        z_value = (n_runs - mu_r) / sigma_r
        p_value = (1 - stats.norm.cdf(z_value)) * 2
        if (p_value > 1):
            p_value = (stats.norm.cdf(z_value)) * 2
        flag = p_value < alpha
        des = f"""======= Runs Test =======
(n1 ({n1}) or n2 ({n2}) > 20)
Runs = {n_runs}
runs_exp (mu_r) = {mu_r:.4f}
std (sigma_r) = {sigma_r:4f}

z-value (observed statistic) = {z_value:.4f}
p-value = {p_value:.4f}
Reject H_0 (Randomness does not exist) → {flag}
"""
        print(des)

        return z_value, p_value



from matplotlib import pyplot as plt
import seaborn as sns
import pandas as pd
import numpy as np
import scipy.stats as stats
import statsmodels.api as sm
import statsmodels.stats.api as sms
import statsmodels.formula.api as smf
import statsmodels.stats.multicomp as smm
import statsmodels.stats.outliers_influence as sso
import statsmodels
import statistics
import math
import time
import itertools
from scipy.optimize import curve_fit
from statsmodels.tsa.ar_model import AutoReg, ar_select_order
from statsmodels.tsa.api import acf, pacf, graphics

# compute p value from t statistics
def tpv (stat, dof, tail):
    if (tail == 'r'):
        return 1 - stats.t.cdf(stat, df = dof) #right
    elif (tail == 'l'):
        return stats.t.cdf(stat, df = dof) #left
    elif (tail == 'db'):
        if(stats.t.cdf(stat, df = dof) > 0.5):
            return 2 * (1 - stats.t.cdf(stat, df = dof)) # double
        else:
            return 2 * stats.t.cdf(stat, df = dof) # double
    else:
        return -1 # error

# p value interpretation
def pvSig (pv):
    print("\n====== p value significance ======")
    if (pv <= 0.01):
        print(">>> highly sig, overwhelming evidence\n    sig, strong evidence\n    not sig, weak evidence\n    not sig, little to no evidence")
    elif (pv <= 0.05 and pv > 0.01):
        print("    highly sig, overwhelming evidence\n>>> sig, strong evidence\n    not sig, weak evidence\n    not sig, little to no evidence")
    elif (pv <= 0.1 and pv > 0.05):
        print("    highly sig, overwhelming evidence\n    sig, strong evidence\n>>> not sig, weak evidence\n    not sig, little to no evidence")
    elif (pv > 0.1):
        print("    highly sig, overwhelming evidence\n    sig, strong evidence\n    not sig, weak evidence\n>>> not sig, little to no evidence")
    else:
        print("BAD INPUT")
    print("===================================\n")

# r value interpretation
def rvInter (rv):
    print("\n====== R value interpretation ======")
    if (rv > 0):
        print("            [positive]")
    elif (rv <0):
        print("            [negative]")
    else:
        print("        [no linear rellation]")
        return

    if (abs(rv) <= 0.25):
        print("    very strong\n    moderately strong\n    moderately weak\n>>> very weak")
    elif (abs(rv) <= 0.5 and abs(rv) > 0.25):
        print("    very strong\n    moderately strong\n>>> moderately weak\n    very weak")
    elif (abs(rv) <= 0.75 and abs(rv) > 0.5):
        print("    very strong\n>>> moderately strong\n    moderately weak\n    very weak")
    elif (abs(rv) <= 1 and abs(rv) > 0.75):
        print(">>> very strong\n    moderately strong\n    moderately weak\n    very weak")
    else:
        print("BAD INPUT")
    print("====================================\n")

def simple_regplot(df, xname, yname):
    _ = sns.regplot(x = xname, y = yname, data = df, color = 'b', ci = None)
    plt.title('Scatter Plot')
    plt.xlabel(xname)
    plt.ylabel(yname)
    plt.show()

def simple_regmod(df, xname, yname):
    # Fit regression model 
    result1 = smf.ols(yname + '~ ' + xname, data = df).fit()
    # Inspect the results
    print(result1.summary())

    b1_1 = result1.params[1]
    b0_1 = result1.params[0]
    print(f"Estimated model: y = {b0_1:.4f} + {b1_1:.4f} x")

def simple_durbin_watson(df, xname, yname, alpha = 0.05):
    print("\n\n========== Durbin-Watson ==========\n")
    
    y_data = df[yname]    
    x_data_ar = []
    x_data_ar = np.asarray(df[xname])
    
    x_data_T = x_data_ar.T
    x_data = pd.DataFrame({xname:x_data_T})
    x_data2 = sm.add_constant(x_data)
    olsmod = sm.OLS(y_data, x_data2)
    result = olsmod.fit()
    
    st, data, ss2 = sso.summary_table(result, alpha = alpha)
    print("Columns in data are: %s" % ss2)
    # Predicted value
    y_pre = data[:, 2]
    # Studentized Residual
    SD = data[:, 10]
    
    x_square_sum = np.vdot(SD, SD)
    print("x_square_sum = ", x_square_sum)
    size = SD.size
    print("size = ", size)
    x_d = np.zeros((size))
    print("x_d = ", x_d)
    l_size = size - 1
    for i in range(l_size):
        x_d[i + 1] = SD[i + 1] - SD[i]
    print("x_d = ", x_d)
    d = np.vdot(x_d, x_d) / x_square_sum
    print("d = ", d)

def chi2_normtest (stand_res, N, alpha = 0.05): 
    m = np.mean(stand_res)
    s = np.std(stand_res)
    prob_bins = np.zeros((N + 1))
    z_bins = np.zeros((N + 1))
    z_bins[0] = -4
    z_bins[N] = 4
    for i in range(0, N+1):
        prob_bins[i] = i/N
    for j in range(1, N):
        z_bins[j] = m + stats.norm.isf(1 - prob_bins[j]) * s
    counts, bins = np.histogram(stand_res, bins=z_bins)
    nobs = counts.sum()
    prob_e = np.zeros((N))
    for i in range(1, N+1):
        prob_e[i - 1] = prob_bins[i] - prob_bins[i - 1]   
    freq_e = nobs * prob_e
    freq_o = counts
    if np.sum(freq_e < 5) > 0:
        print("Rule of five is not met.")
    else:
        chi_stat, pval = stats.chisquare(freq_o, freq_e)
        chi_pval = stats.chi2.sf(chi_stat, N - 3)
        print("Chi-squared test: statistics = %0.4f, p-value = %0.4f" % (chi_stat, chi_pval))
    df = freq_o.shape[0]-3
    crit_value = stats.chi2.ppf(1 - alpha, df)
    print("Critical value = %0.4f (defree of freedom = %d)" % (crit_value, df))
    
    return chi_pval

def runsTest(l, l_median): 
    runs, n1, n2 = 1, 0, 0
    if(l[0]) >= l_median:
        n1 += 1   
    else:
        n2 += 1     
    # Checking for start of new run 
    for i in range(len(l)): 
        # no. of runs 
        if (l[i] >= l_median and l[i-1] < l_median) or (l[i] < l_median and l[i-1] >= l_median):
            runs += 1  
            # print(i, runs)
        # no. of positive values 
        if(l[i]) >= l_median: 
            n1 += 1   
        # no. of negative values 
        else: 
            n2 += 1   
    runs_exp = ((2*n1*n2)/(n1+n2)) + 1
    stan_dev = math.sqrt((2*n1*n2*(2*n1*n2-n1-n2))/(((n1+n2)**2)*(n1+n2-1))) 
    z = (runs-runs_exp)/stan_dev
    pval_z = stats.norm.sf(abs(z)) * 2
    print('runs = ', runs)
    print('n1 = ', n1)
    print('n2 = ', n2)
    print('runs_exp = ', runs_exp)
    print('stan_dev = ', stan_dev)
    print('z = ', z)
    print('pval_z = ', pval_z)
    return pval_z

def simple_residual(df, xname, yname, alpha = 0.05, resd_all = False, nobins = 6):
    print("\n\n----------------------------\n|Residual Analysis - simple|\n----------------------------\n")
    print("using alpha = ", alpha)
    print("\n\n ◆ Residuals\n")
    
    # Fit regression model 
    result = smf.ols(yname + '~' + xname, data = df).fit()

    # studentized residual
    st1, data1, ss3 = sso.summary_table(result, alpha = alpha)
    Residual = data1[:, 8]
    STD_Residual = data1[:,10]
    mu = np.mean(STD_Residual)
    sigma = np.std(STD_Residual)
    
    if(resd_all == True):
        print("Original Residuals: \n", Residual, "\n")
        print("Standardized Residuals: \n", STD_Residual, "\n")
        print("mu:", mu)
        print("sigma:", sigma)
    else:
        print("mu:", mu)
        print("sigma:", sigma)
    
    # Normality Test
    print("\n\n ◆ Error Normality Test\n")
    print("H0: Errors are normally distributed.")
    print("H1: Errors are not normally distributed.")
    
    # Histogram
    print("\n\n   ◇ Histogram\n")
    counts, bins, patches = plt.hist(STD_Residual, nobins, density = False, facecolor = 'black', alpha = 0.75)

    plt.xlabel('Standardized Residuals')
    plt.ylabel('Frequency')
    plt.title('Histogram of Standardized Residuals')
    plt.grid(True)
    bin_centers = [np.mean(k) for k in zip(bins[:-1], bins[1:])]
    plt.show()

    print(counts)
    print(bins)
    
    # Shapiro Test
    print("\n\n   ◇ Shapiro Test\n")
    stat, spv = stats.shapiro(STD_Residual)
    print(f"Statistics = {stat:.4f}, p-value = {spv:.4f}")
    pvSig(spv)
    
    # Chi^2 Test
    print("\n\n   ◇ Chi-squared Test\n")
    stand_res = STD_Residual
    N = nobins
    
    m = np.mean(stand_res)
    s = np.std(stand_res)
    prob_bins = np.zeros((N + 1))
    z_bins = np.zeros((N + 1))
    z_bins[0] = -4
    z_bins[N] = 4
    for i in range(0, N+1):
        prob_bins[i] = i/N
    for j in range(1, N):
        z_bins[j] = m + stats.norm.isf(1 - prob_bins[j]) * s
    counts, bins = np.histogram(stand_res, bins=z_bins)
    nobs = counts.sum()
    prob_e = np.zeros((N))
    for i in range(1, N+1):
        prob_e[i - 1] = prob_bins[i] - prob_bins[i - 1]   
    freq_e = nobs * prob_e
    freq_o = counts
    if np.sum(freq_e < 5) > 0:
        print("Rule of five is not met.")
    else:
        chi_stat, pval = stats.chisquare(freq_o, freq_e)
        chi_pval = stats.chi2.sf(chi_stat, N - 3)
        print("Chi-squared test: statistics = %0.4f, p-value = %0.4f" % (chi_stat, chi_pval))
    df_fq = freq_o.shape[0]-3
    crit_value = stats.chi2.ppf(1 - alpha, df_fq)
    print("Critical value = %0.4f (defree of freedom = %d)" % (crit_value, df_fq))

    #pvSig(chi_pval)
    
    # Homoscedasticity and Heteroscedasticity 
    print("\n\n ◆ Homoscedasticity and Heteroscedasticity\n")
    print("H_0: Randomness exists")
    print("H_1: Randomness doesn't exist")
    Id1 = data1[:, 0]
    plt.plot(Id1, STD_Residual, 'o', color = 'gray')
    plt.axhline(y=0, color = 'blue')
    plt.axhline(y=2, color = 'red')
    plt.axhline(y=-2, color = 'red')
    plt.title('Standardized Residual Plot')
    plt.xlabel('Observation No.')
    plt.ylabel('Standardized Residual')
    plt.show()

    # Dependence of the Error Variable
    print("\n\n ◆ Dependence of the Error Variable (Run Test)\n")
    print("H_0: Sample is random")
    print("H_1: Sample is not random")
    SD_median = statistics.median(STD_Residual)
    Z_pval = runsTest(STD_Residual, SD_median) 
    print('p-value for run test z-statistic= ', Z_pval)
    pvSig(Z_pval)
    
    # Outliers
    print("\n\n ◆ Outliers Finding\n")
    print("(remove by yourself!)\n")
    df_out = pd.DataFrame(STD_Residual, columns = ['SD'])
    filter = (df_out['SD'] < -2) | (df_out['SD'] > 2)
    print("Outliers by SD = ")
    print(df_out['SD'].loc[filter])
    print("\nActual ID: ", df_out['SD'].loc[filter].index+1)
    
    # Influential Observations
    print("\n\n ◆ Influential observations Finding\n")
    x_data = df[xname].values
    y_data = df[yname].values
    cov_mat1 = np.cov(y_data, x_data)
    x_data_bar = x_data.mean()
    data_nobs = len(x_data)
    h_val = 1 / data_nobs + (x_data - x_data_bar) ** 2 / (data_nobs - 1) / cov_mat1[1,1]
    # print(h_val)
    df_hi = pd.DataFrame(h_val, columns = ['hi'])
    filter = (df_hi['hi'] > nobins / data_nobs )
    print("Influential Observations by hi = ", df_hi['hi'].loc[filter])
    print("\nAutal ID: ", df_hi['hi'].loc[filter].index+1)

def simple_modass(df, xname, yname, alpha = 0.05, tail = 'db'):
    # Fit regression model 
    result1 = smf.ols(yname + '~ ' + xname, data = df).fit()

    b1_1 = result1.params[1]
    b0_1 = result1.params[0]
    print(f"Estimated model: y = {b0_1:.4f} + {b1_1:.4f} x")

    print("\n\n---------------------------\n|     Model Assessing     |\n---------------------------\n")
    print("using alpha = ", alpha)

    print("\n\n ◆ Standard Error of Estimate\n")
    s2_e = result1.mse_resid
    print(f"MSE = {s2_e:f}")
    s_e = result1.mse_resid ** 0.5
    print(f"Standard errors = {s_e:f}")
    y_bar = df[yname].mean()
    print(f"y mean = {y_bar:.4f}")
    print(f"The absolute value of standard errors is about {abs(s_e/y_bar)*100:.0f}% of mean of independent variables.\n")

    print("\n\n ◆ Coefficient of Determination\n")
    R2 = result1.rsquared
    print(f"R^2 = {R2:f}")
    R = np.sign(b1_1) * R2 ** 0.5
    print(f"R = {R:f}")

    print(f"\nR^2 value interpretation\nAbout {R2*100:.0f}% of the variation in the dependent variables is explained by independent ones, the rest remains unexplained.")
    rvInter(R)

    print("\n\n ◆ Studetn-t test for beta1(slope)\n")
    dof = len(df) - 2
    tv = R * ((dof - 2)/(1 - R ** 2)) ** 0.5
    LCL = stats.t.ppf(alpha / 2, dof - 2)
    UCL = stats.t.ppf(1 - alpha / 2, dof - 2)
    print('t = ', tv)
    print('t_LCL = ', LCL)
    print('t_UCL = ', UCL)

    print(f"\np-value of t-stat tail: {tail}")
    tp = tpv(tv, dof, tail)
    print("p-value of t test = ", tp)
    pvSig(tp)
    
    print("\n\n ◆ Coefficient of Correlation\n")
    cor_mat = np.corrcoef(df[[xname, yname]].values, rowvar = False)
    n = df.shape[0]
    r = cor_mat[1,0]
    tv_cc = r * ((n-2)/(1 - r**2)) ** 0.5
    t_critical = stats.t.ppf(0.975, n - 2)
    pval = stats.t.sf(np.abs(tv_cc), n - 2)*2
    
    print('r = ', r)
    print('t_critical = ', t_critical)
    print('t = ', tv_cc)
    print('p_value = ', pval)

def simple_CIPIPRE (x, y, x1, alpha = 0.05):
    print("\n\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n|CI PI for simple regression|\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n")
    print("using alpha = ", alpha)
    
    x_new = np.array([1, x1])
    print("make Confidence Interval and Prediction Interval predictions at mean of x = ", x1)
    x2 = sm.add_constant(x)
    olsmod = sm.OLS(y, x2)
    result_reg = olsmod.fit()
    y_head = np.dot(result_reg.params, x_new)
    print("y_head = ", y_head)
    (t_minus, t_plus) = stats.t.interval(alpha = (1.0 - alpha), df = result_reg.df_resid )
    cov_mat1 = np.cov(y, x)
    x_bar = x.mean()
    core1 = (1 / result_reg.nobs + (x1 - x_bar) ** 2 / (result_reg.nobs - 1) / cov_mat1[1,1] ) ** 0.5
    core2 = (1 + 1 / result_reg.nobs + (x1 - x_bar) ** 2 / (result_reg.nobs - 1) / cov_mat1[1,1] ) ** 0.5
    lower_bound = y_head + t_minus * (result_reg.mse_resid ** 0.5) * core1
    upper_bound = y_head + t_plus * (result_reg.mse_resid ** 0.5) * core1
    half_interval = t_plus * (result_reg.mse_resid ** 0.5) * core1
    lower_bound2 = y_head + t_minus * (result_reg.mse_resid ** 0.5) * core2
    upper_bound2 = y_head + t_plus * (result_reg.mse_resid ** 0.5) * core2
    half_interval2 = t_plus * (result_reg.mse_resid ** 0.5) * core2
    
    print(f"\n{100*(1-alpha):.0f}% confidence interval for mean: [{lower_bound:.4f}, {upper_bound:.4f}], or {y_head:.4f} +- {half_interval:.4f}")
    print(f"\n{100*(1-alpha):.0f}% prediction interval: [{lower_bound2:.4f}, {upper_bound2:.4f}], or {y_head:.4f} +- {half_interval2:.4f}")

def simple_CIPIINT_regplot(df, xname, yname, alpha = 0.05):
    print("\n\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n|CI PI Interval plot - simple|\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n")
    print("using alpha = ", alpha)
    
    df_sorted = df.sort_values([xname])
    result = smf.ols(yname + '~' + xname, data = df_sorted).fit()
    x = df_sorted[xname].values
    y = df_sorted[yname].values
    st, data, ss2 = sso.summary_table(result, alpha = alpha)
    fittedvalues = data[:, 2]
    predict_mean_se  = data[:, 3]
    predict_mean_ci_low, predict_mean_ci_upp = data[:, 4:6].T
    predict_ci_low, predict_ci_upp = data[:, 6:8].T
    
    plt.plot(x, y, 'o', color = 'gray')
    plt.plot(x, fittedvalues, '-', lw=0.5)
    plt.plot(x, predict_mean_ci_low, 'r-', lw=0.4)
    plt.plot(x, predict_mean_ci_upp, 'r-', lw=0.4)
    plt.plot(x, predict_ci_low, 'b--', lw=0.4)
    plt.plot(x, predict_ci_upp, 'b--', lw=0.4)
    plt.title('CI PI plot')
    plt.xlabel(xname)
    plt.ylabel(yname)
    plt.legend(['data points', 'regression model', 'confidence interval', 'prediction interval'], title = 'Legends', bbox_to_anchor = (1.3, 1), prop={'size': 6})
    plt.show()

def simple(step, df, xname, yname, alpha = 0.05, tail = 'db', nobins = 6, resd_all = False):
    if step == 1:
        simple_regplot(df, xname, yname)
    elif step == 2:
        print("\npropose a statistical model\n")
    elif step == 3:
        simple_regmod(df, xname, yname)
    elif step == 4:
        print("\nfor autocorrelation and others, please determine by yourself!\n")
        simple_durbin_watson(df, xname, yname, alpha = alpha)
    elif step == 5:
        print("\nremember to remove outliers or do some modifications.\n")
        simple_residual(df, xname, yname, alpha = alpha, resd_all = resd_all, nobins = nobins)
    elif step == 6:
        simple_modass(df, xname, yname, alpha = alpha, tail = tail)
    elif step == 7:
        print("\ninterpretation\n")
    elif step == 8:
        print("\nsimple_CIPIPRE(df[xname], df[yname], x_input...) won't run here\n")
        simple_CIPIINT_regplot(df, xname, yname, alpha = alpha)
    else:
        print("\nbad input for step!\n")

def multiple_regplot(df, xnames, yname):
    for aname in xnames:
        x_var = df[aname].values
        _ = sns.regplot(x = x_var, y = df[yname].values, data = df, color = 'b', ci = None)
        plt.xlabel(aname)
        plt.ylabel(yname)
        plt.show()

    df_ = df[[yname] + xnames]
    corr1 = df_.corr()
    corr1
    _ = sns.heatmap(corr1, annot = True)

def multiple_modpropose(xnames, yname):
    print("\n\n---------- Model Proposal ----------\n")
    print("Model proposal,<br>")
    
    mod = "$y = \\beta_0 + "
    for i in range(len(xnames)):
        coe = "\\beta_" + str(i+1) + "x_" + str(i+1) + " + "
        mod = mod + coe
    mod = mod + "\\epsilon$<br>"
    print(mod)
    
    print("where y is ", yname, "<br>")
    
    exp = "and "
    for j in range(len(xnames)):
        xexp = "$x_" + str(j+1) + "$ is " + xnames[j] + ", "
        exp = exp + xexp
    
    print(exp)

def multiple_regmod(df, xnames, yname, alpha = 0.05):
    y_data = df[yname]
    x_data_ar = []
    for i in range(len(xnames)):
        x_data_ar.append(df[xnames[i]])
    x_data_ar = np.asarray(x_data_ar)
    
    x_data_T = x_data_ar.T
    x_data = pd.DataFrame(x_data_T, columns = xnames)
    x_data2 = sm.add_constant(x_data)
    olsmod = sm.OLS(y_data, x_data2)
    result = olsmod.fit()
    print(f"\n\nusing alpha = {alpha:.2f}")
    print(result.summary())
    
    print("\nEstimated model: y = %0.4f" % (result.params[0]), end = ' ')
    for c, x in zip(result.params[1:], list(range(1,len(xnames)+1))):
        print('+', end = '') if c> 0 else print('-', end = '')
        print(" %0.4f x%d " % (abs(c), x), end = '') 

def multiple_durbin_watson(df, xnames, yname, alpha = 0.05):
    print("\n\n========== Durbin-Watson ==========\n")
    
    y_data = df[yname]    
    x_data_ar = []
    for i in range(len(xnames)):
        x_data_ar.append(df[xnames[i]])
    x_data_ar = np.asarray(x_data_ar)
    
    x_data_T = x_data_ar.T
    x_data = pd.DataFrame(x_data_T, columns = xnames)
    x_data2 = sm.add_constant(x_data)
    olsmod = sm.OLS(y_data, x_data2)
    result = olsmod.fit()
    
    st, data, ss2 = sso.summary_table(result, alpha = alpha)
    print("Columns in data are: %s" % ss2)
    # Predicted value
    y_pre = data[:, 2]
    # Studentized Residual
    SD = data[:, 10]
    
    x_square_sum = np.vdot(SD, SD)
    print("x_square_sum = ", x_square_sum)
    size = SD.size
    print("size = ", size)
    x_d = np.zeros((size))
    print("x_d = ", x_d)
    l_size = size - 1
    for i in range(l_size):
        x_d[i + 1] = SD[i + 1] - SD[i]
    print("x_d = ", x_d)
    d = np.vdot(x_d, x_d) / x_square_sum
    print("d = ", d)

def multiple_residual(df, xnames, yname, alpha = 0.05, nobins = 6):
    print("\n\n----------------------------\n|Residual Analysis - multiple|\n----------------------------\n")
    print("using alpha = ", alpha)
    print("\n\n ◆ Residuals\n")
    
    y_data = df[yname]    
    x_data_ar = []
    for i in range(len(xnames)):
        x_data_ar.append(df[xnames[i]])
    x_data_ar = np.asarray(x_data_ar)
    
    x_data_T = x_data_ar.T
    x_data = pd.DataFrame(x_data_T, columns = xnames)
    x_data2 = sm.add_constant(x_data)
    olsmod = sm.OLS(y_data, x_data2)
    result = olsmod.fit()
    
    st, data, ss2 = sso.summary_table(result, alpha = alpha)
    print("Columns in data are: %s" % ss2)
    # Predicted value
    y_pre = data[:, 2]
    # Studentized Residual
    SD = data[:, 10]
    
    mu = np.mean(SD)
    sigma = np.std(SD)
    
    # Normality Test
    print("\n\n ◆ Error Normality Test\n")
    print("H0: Errors are normally distributed.")
    print("H1: Errors are not normally distributed.")

    # Histogram
    print("\n\n   ◇ Histogram\n")
    fig, ax = plt.subplots()
    counts, bins, patches = plt.hist(SD, nobins, density=False, facecolor='g', alpha=0.75)
    plt.xlabel('Standardized Residuals')
    plt.ylabel('Frequency')
    plt.title('Histogram of Standardized Residuals_Car Prices')
    plt.grid(True)
    bin_centers = [np.mean(k) for k in zip(bins[:-1], bins[1:])]
    plt.show()
    
    print(counts)
    print(bins)

    # qqplot
    print("\n\n   ◇ QQ-plot\n")
    fig = sm.qqplot(SD, stats.norm, fit = True, line = '45')
    plt.show()
    print()
    
    # Shapiro Test
    print("\n\n   ◇ Shapiro Test\n")
    stat, spv = stats.shapiro(SD)
    print(f"Statistics = {stat:.4f}, p-value = {spv:.4f}")
    pvSig(spv)

    # Chi^2 Test
    print("\n\n   ◇ Chi-squared Test\n")
    stand_res = SD
    N = nobins - 1
    
    m = np.mean(stand_res)
    s = np.std(stand_res)
    prob_bins = np.zeros((N + 1))
    z_bins = np.zeros((N + 1))
    z_bins[0] = -4
    z_bins[N] = 4
    for i in range(0, N+1):
        prob_bins[i] = i/N
    for j in range(1, N):
        z_bins[j] = m + stats.norm.isf(1 - prob_bins[j]) * s
    counts, bins = np.histogram(stand_res, bins=z_bins)
    nobs = counts.sum()
    prob_e = np.zeros((N))
    for i in range(1, N+1):
        prob_e[i - 1] = prob_bins[i] - prob_bins[i - 1]   
    freq_e = nobs * prob_e
    freq_o = counts
    if np.sum(freq_e < 5) > 0:
        print("Rule of five is not met.")
    else:
        chi_stat, pval = stats.chisquare(freq_o, freq_e)
        chi_pval = stats.chi2.sf(chi_stat, N - 3)
        print("Chi-squared test: statistics = %0.4f, p-value = %0.4f" % (chi_stat, chi_pval))
    df_fq = freq_o.shape[0]-3
    crit_value = stats.chi2.ppf(1 - alpha, df_fq)
    print("Critical value = %0.4f (defree of freedom = %d)" % (crit_value, df_fq))

    #pvSig(chi_pval)
    
    # Homoscedasticity and Heteroscedasticity 
    print("\n\n ◆ Homoscedasticity and Heteroscedasticity\n")
    print("H_0:Randomness exists")
    print("H_0:Randomness doesn't exist")
    
    st, data, ss2 = sso.summary_table(result, alpha = alpha)
    print("\nColumns in data are: %s" % ss2)
        # Predicted value
    y_pre = data[:, 2]
        # Studentized Residual
    SD = data[:, 10]
    
    plt.plot(y_pre, SD, 'o', color = 'gray')
    plt.axhline(y=2, color = 'red', lw = 0.8)
    plt.axhline(y=0, color = 'blue')
    plt.axhline(y=-2, color = 'red', lw = 0.8)
    plt.title('Standardized Residual Plot')
    plt.xlabel('Predicted y value')
    plt.ylabel('Standardized Residual')
    plt.show()
    
    ##### autocorrelation
    # Dependence of the Error Variable
    print("\n\n ◆ Dependence of the Error Variable (Run Test)\n")
    print("H_0: Sample is random")
    print("H_1: Sample is not random")

    print("\nColumns in data are: %s" % ss2)
    
    Id1 = data[:, 0]
    plt.plot(Id1, SD, 'o', color = 'gray')
    plt.axhline(y=0, color = 'blue')
    plt.axhline(y=2, color = 'red')
    plt.axhline(y=-2, color = 'red')
    plt.title('Standardized Residual Plot')
    plt.xlabel('Observation No.')
    plt.ylabel('Standardized Residual')
    plt.show()

    SD_median = statistics.median(SD)
    Z_pval = runsTest(SD, SD_median)
    print('p_value for Z-statistic= ', Z_pval)
    
    pvSig(Z_pval)
    
    # Outliers
    print("\n\n ◆ Outliers Finding\n")
    df_out = pd.DataFrame(SD, columns = ['SD'])
    filter = (df_out['SD'] < -2) | (df_out['SD'] > 2)
    print("Outliers by SD = ")
    print(df_out['SD'].loc[filter])
    print("\nActual ID: ", df_out['SD'].loc[filter].index+1)
    
    
    ## Influential Observations by hii
    print("\n\n ◆ Influential observations Finding by hii\n")
    x_data2 = np.array(x_data2)
    H = np.matmul(x_data2, np.linalg.solve(np.matmul(x_data2.T, x_data2), x_data2.T))
    df['hii'] = np.diagonal(H)
    df_1h = pd.DataFrame(df['hii'])
    k = result.df_model
    n = len(df_1h['hii'])
    h_level = 3 * (k+1) / n 
    print("h_level = ", h_level)
    filter = (df_1h['hii'] > h_level)
    print("\nInfluential Observations by hi =\n")
    print(df_1h['hii'].loc[filter])
    
    # Influential Observations by Cook's Distance
    print("\n\n ◆ Influential observations Finding by Cook's Distance\n")
    s2_e = result.mse_resid
    k = result.df_model
    y_a = data[:, 1]
    y_f = data[:, 2]
    h_i = df['hii']
    CD_arr = np.square(y_a - y_f) / s2_e / (k - 1) * h_i / np.square(1 - h_i)
    CD = np.array(CD_arr)
    df_cd = pd.DataFrame(CD,columns = ['CD'])
    print(df_cd.head())
    filter = (df_cd['CD'] > 1 )
    print("Influential Observations by Cook's Distances =\n")
    print(df_cd['CD'].loc[filter])

def multiple_modass(df, xnames, yname, alpha = 0.05):
    y_data = df[yname]
    x_data_ar = []
    for i in range(len(xnames)):
        x_data_ar.append(df[xnames[i]])
    x_data_ar = np.asarray(x_data_ar)
    
    x_data_T = x_data_ar.T
    x_data = pd.DataFrame(x_data_T, columns = xnames)
    x_data2 = sm.add_constant(x_data)
    olsmod = sm.OLS(y_data, x_data2)
    result = olsmod.fit()
    
    print("\n\n---------------------------\n|     Model Assessing     |\n---------------------------\n")
    print("using alpha = ", alpha)

    print("\n\n ◆ Standard Error of Estimate\n")
    s2_e = result.mse_resid
    print(f"MSE = {s2_e:f}")
    s_e = result.mse_resid ** 0.5
    print("Standard error = ", s_e)
    y_bar = df[yname].mean()
    print("y mean = ", y_bar)
    print("y STD = ", df[yname].std())
    print(f"The absolute value of standard errors is about {abs(s_e/y_bar)*100:.0f}% of mean of independent variables.\n")

    R2 = result.rsquared
    print("\nCoefficient of Determination")
    print("R^2 = ", result.rsquared)
    print("Adjusted R^2 = ", result.rsquared_adj)

    print(f"\nR^2 value interpretation\nAbout {R2*100:.0f}% of the variation in the dependent variables is explained by the model, the rest remains unexplained.")
    rvInter(R2**0.5)

    print("\n\n ◆ Over-fitting?\n")
    diffrra = abs(result.rsquared - result.rsquared_adj)
    print("|R^2 - Ra^2| = ", diffrra)
    if(diffrra > 0.06):
        print("|R^2 - Ra^2| >= 0.06 indicating that the model has the problem of over-fitting.")
    else:
        print("|R^2 - Ra^2| < 0.06 indicating that the model doesn't have the problem of over-fitting.")

    print("\n\n ◆ F-test of ANOVA\n")
    print("Testing hypothesis,")
    print("H_0: \beta_1 = \beta_2 = \dots = \beta_n = 0<br>")
    print("H_1: \text{at least one } \beta_i \neq 0")

    f_res = result.fvalue
    MSE = result.mse_resid
    df_model = result.df_model
    df_error = result.df_resid
    MSR = f_res * MSE
    SSR = MSR * df_model
    print("SSR = ", SSR, "\tdf = ", df_model, "\tMSR = ", MSR)
    print("SSE = ", MSE * df_error, "\tdf = ", df_error, "\tMSE = ", MSE)
    print("F = MSR / MSE = ", MSR / MSE)
    fpv = result.f_pvalue
    print("F p-value = ", fpv)

    pvSig(fpv)

def multiple_CIPIPRE_ (xdata, yval, x1, alpha = 0.05):
    print("\n\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n|CI PI for simple regression|\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n")
    print("using alpha = ", alpha)
    
    print("To make Confidence Interval and Prediction Interval prediction at mean of x = ", x1)
    x_data_T = xdata.T
    x_data2 = sm.add_constant(x_data_T)
    olsmod = sm.OLS(yval, x_data2)
    result_reg = olsmod.fit()
    y_head = np.dot(result_reg.params, x1)
    print("y_head = ", y_head)
    (t_minus, t_plus) = stats.t.interval(alpha = (1.0 - alpha), df = result_reg.df_resid )
    core1 = (result_reg.mse_resid * np.matmul(x1, np.linalg.solve(np.matmul(x_data2.T, x_data2), x1))) ** 0.5
    lower_bound = y_head +  t_minus * core1
    upper_bound = y_head +  t_plus * core1
    core2 = (result_reg.mse_resid * (1 + np.matmul(x1, np.linalg.solve(np.matmul(x_data2.T, x_data2), x1)))) ** 0.5
    lower_bound2 = y_head +  t_minus * core2
    upper_bound2 = y_head +  t_plus * core2
    
    print(f"\n{100*(1-alpha):.0f}% confidence interval for mean: [{lower_bound:.4f}, {upper_bound:.4f}]")
    print(f"\n{100*(1-alpha):.0f}% prediction interval: [{lower_bound2:.4f}, {upper_bound2:.4f}]")

def multiple_CIPIPRE (df, xnames, yname, xx, alpha = 0.05):
    x0 = [1]
    x1 = x0 + xx

    yval = df[yname]
    xdata_ar = []

    for i in range(len(xnames)):
        xdata_ar.append(df[xnames[i]])
    xdata_ar = np.asarray(xdata_ar)

    x1_ = np.array(x1)

    multiple_CIPIPRE_ (xdata_ar, yval, x1_, alpha)

def multiple(step, df, xnames, yname, alpha = 0.05, tail = 'db', nobins = 6):
    if step == 1:
        multiple_regplot(df, xnames, yname)
    elif step == 2:
        multiple_modpropose(xnames, yname)
    elif step == 3:
        multiple_regmod(df, xnames, yname)
    elif step == 4:
        print("\nfor autocorrelation and others, please determine by yourself!\n")
        multiple_durbin_watson(df, xnames, yname, alpha = alpha)
    elif step == 5:
        print("\nremember to remove outliers or do some modifications.\n")
        multiple_residual(df, xnames, yname, alpha = alpha, nobins = nobins)
    elif step == 6:
        multiple_modass(df, xnames, yname, alpha = alpha)
    elif step == 7:
        print("\ninterpretation\n")
    elif step == 8:
        print("\multiple_CIPIPRE (df, xnames, yname, xx...) won't run here\n")
    else:
        print("\nbad input for step!\n")

def time_add(df, name = 'Time'):
    time = []
    for i in range(df.shape[0]):
        time.append(i)
    print(time)
    
    df[name] = time
    return df

def value_map_ln(df, target):
    lnv = []
    for i in range(df.shape[0]):
        lnv.append(math.log((df[target].values[i])))
    
    newname = "ln_" + target
    df[newname] = lnv
    return df

def outliers_rm(df, out):
    df = df.drop(df.index[out])
    df = df.reset_index()
    return df



from numpy.testing._private.utils import measure
import statsmodels.stats.outliers_influence as sso
import statsmodels.api as sm
import pandas as pd
import numpy as np


def cma(y_v, period: int):
    y_v_MA_a = np.zeros(len(y_v))
    y_v_MA_a[:] = np.nan
    n = period
    if n % 2 == 0:  # even
        halfwin = int(n / 2)
        y_v_MA_ta = np.zeros(len(y_v))
        # first attempt
        for i in range(halfwin, len(y_v) - halfwin + 1):
            y_v_MA_ta[i] = np.mean(y_v[(i-halfwin): (i+halfwin)])
        # second attempt
        for ii in range(halfwin, len(y_v) - halfwin):
            y_v_MA_a[ii] = np.mean(y_v_MA_ta[(ii): (ii+2)])
    else:  # odd
        halfwin = int((n - 1) / 2)
        for i in range(halfwin, len(y_v) - halfwin):
            y_v_MA_a[i] = np.mean(y_v[(i-halfwin): (i+halfwin+1)])
    return y_v_MA_a


def seasonal_index(y_v, period: int, show=False, option='cma'):
    """
    return seasonal index and a Series with all observations linked to its seasonal index
    """
    n = period
    if option == 'cma':
        SI_MA_a = np.zeros(len(y_v))
        SI_MA_a[:] = np.nan
        SI_MA_a = y_v / cma(y_v, period)
        SI_id_s = np.arange(1, len(y_v)+1)
        SI_id = SI_id_s - np.floor(SI_id_s / n) * n
        SI_id[np.where((SI_id[:] == 0))] = n
        SI_MA_df = pd.DataFrame({'SIMA': SI_MA_a, 'SIid': SI_id})
        SI_MA_u = np.zeros(n)
        for j in range(1, n+1):
            SI_MA_u[j-1] = SI_MA_df['SIMA'][SI_MA_df['SIid']
                                            == j].dropna().mean()
        SI_MA = SI_MA_u / sum(SI_MA_u) * n
        if show:
            print('Seasonal Index:', SI_MA)
        return SI_MA, SI_MA_df['SIid']
    elif option == 'lr':
        y_data = y_v
        X_data_ar = np.arange(1, len(y_v)+1)
        X_data_T = X_data_ar.T
        X_data = pd.DataFrame(X_data_T, columns=['Time'])
        X_data = sm.add_constant(X_data)
        olsmod = sm.OLS(y_data, X_data)
        result_reg = olsmod.fit()
        st, data, ss2 = sso.summary_table(result_reg, alpha=0.05)
        y_v_LR_a = data[:, 2]
        SI_LR_a = y_v / y_v_LR_a
        SI_id_s = np.arange(1, len(y_v)+1)
        SI_id = SI_id_s - np.floor(SI_id_s / n) * n
        SI_id[np.where((SI_id[:] == 0))] = n
        SI_LR_a_df = pd.DataFrame({'SILR': SI_LR_a, 'SIid': SI_id})
        SI_LR_u = np.zeros(n)
        for j in range(1, n+1):
            SI_LR_u[j-1] = SI_LR_a_df['SILR'][SI_LR_a_df['SIid']
                                              == j].dropna().mean()
        SI_LR = SI_LR_u / sum(SI_LR_u) * n

        if show:
            print('Seasonal Index:', SI_LR)

        return SI_LR, SI_LR_a_df['SIid']


def deseasonalize(y_v, period: int, show=False, option='cma'):
    y_v_SI_MA = np.zeros(len(y_v))
    DSI_y_v = np.zeros(len(y_v))
    if option == 'cma':
        SI_MA, SIid = seasonal_index(y_v, period, show, option=option)

        for k in range(0, len(y_v)):
            Idd = int(SIid[k] - 1)
            y_v_SI_MA[k] = SI_MA[Idd]
            DSI_y_v[k] = y_v[k] / SI_MA[Idd]
        if show:
            print('Deseasonalized Data:', DSI_y_v)

        SI_MA_result_m = np.array([SIid, y_v_SI_MA, y_v, DSI_y_v])
        SI_MA_result_df = pd.DataFrame(SI_MA_result_m.T, columns=[
            'SID', 'SeaIdx', 'orig', 'Des_Y'])

        return {'deasonalized_y': DSI_y_v, 'seasonal_i': SIid, 'des_result': SI_MA_result_df, 'SI_MA': SI_MA}
    elif option == 'lr':
        y_v_SI_LR = np.zeros(len(y_v))
        DSI_y_v = np.zeros(len(y_v))
        SI_LR, SIid = seasonal_index(y_v, period, show, option=option)
        for k in range(0, len(y_v)):
            Idd = int(SIid[k] - 1)
            y_v_SI_LR[k] = SI_LR[Idd]
            DSI_y_v[k] = y_v[k] / SI_LR[Idd]

        if show:
            print('Deseasonalized Data:', DSI_y_v)
        SI_LR_result_a = np.array([SIid, y_v_SI_LR, y_v, DSI_y_v])
        SI_LR_result_df = pd.DataFrame(SI_LR_result_a.T, columns=[
                                       'SID', 'SeaIdx', 'orig', 'Des_Y'])
        return {'deasonalized_y': DSI_y_v, 'seasonal_i': SIid, "des_result": SI_LR_result_df, "SI_LR": SI_LR}


def smoothing_cma(df, y_name, time_name, period, show=False):
    DSI_y_v, SIid, des_result, SI_MA = deseasonalize(
        df[y_name], period, show, option='cma').values()
    des_result.rename(columns={'Des_Y': f'Des_{y_name}'}, inplace=True)
    return pd.concat([df, des_result], axis=1)


def smoothing_lr(df, y_name, time_name, period, show=False):
    DSI_y_v, SIid, des_result, SI_LR = deseasonalize(
        df[y_name], period, show, option='lr').values()
    des_result.rename(columns={'Des_Y': f'Des_{y_name}'}, inplace=True)
    return pd.concat([df, des_result], axis=1)


def smoothing(df, y_name, time_name, period, show=False, option='cma'):
    DSI_y_v, SIid, des_result, SI = deseasonalize(
        df[y_name], period, show, option=option).values()
    des_result.rename(columns={'Des_Y': f'Des_{y_name}'}, inplace=True)
    return pd.concat([df, des_result], axis=1)


def seasonal_prediction(df, df_result, y_name, time_name, new_t, period, show=False, option='cma'):
    """
    df should be the deseasoned df if possible
    """
    y_v = df[y_name]
    if len(new_t) == 0:
        _, data, _ = sso.summary_table(df_result, alpha=0.05)
        trend_proj = df_result.predict(sm.add_constant(new_t))
        df_result.predict(sm.add_constant(new_t))
        tdf = df.copy()
        tdf[f'Pre_{y_name}'] = data[:, 2] * tdf['SeaIdx']
        return tdf
    else:
        new_t = np.array(new_t)
        SI, SIid = seasonal_index(y_v, period, show, option=option)
        # des_df = smoothing_cma(df, y_name, time_name,
        #                        period=period, show=show)  # final df secured

        trend_proj = df_result.predict(sm.add_constant(new_t))
        seasonal_adj = trend_proj * SI
        # new_t = np.arange(12, 16)
        _, data, _ = sso.summary_table(df_result, alpha=0.05)
        trend_proj = df_result.predict(sm.add_constant(new_t))
        df_result.predict(sm.add_constant(new_t))
        tdf = df.copy()
        tdf[f'Pre_{y_name}'] = data[:, 2] * tdf['SeaIdx']

        # tdf[x_name] = np.append(tdf[x_name], new_t)
        for i, t in enumerate(new_t):
            tdf = tdf.append({time_name: t, 'SID': tdf['SID'].values[-(1 + i) - (len(new_t) - (1 + i))], 'SeaIdx': tdf['SeaIdx'].values[-(1 + i) - (
                len(new_t) - (1 + i))], f'Pre_{y_name}': trend_proj[i] * tdf['SeaIdx'].values[-(1 + i) - (len(new_t) - (1 + i))]}, ignore_index=True)

    return tdf


def mav(df, n, y_name, time_name, weight: list = None):
    """
    mgt2001.ts.mav
    """
    ma_p_a = np.zeros(df.shape[0] + 1)
    y_v = df[y_name].dropna().values
    ma_p_a[:] = np.nan
    for i in range(n, len(ma_p_a)):
        if weight is not None:
            ma_p_a[i] = np.sum(y_v[i-n: i] * np.array(weight))
        else:
            ma_p_a[i] = np.mean(y_v[i-n: i])
    t1 = np.append(df[time_name].dropna().values[:len(y_v)], [
                   df[time_name].dropna().values[:len(y_v)][-1] + 1])

    ma_df_p = pd.DataFrame(
        {time_name: t1, y_name: np.append(y_v, [np.nan]), f'MA({n})': ma_p_a})
    return pd.concat([ma_df_p, df.drop(columns=[time_name, y_name])], axis=1)


def exp_smoothing(df, alpha, y_name, time_name):
    y_v = df[y_name].dropna().values
    es_df = pd.DataFrame({'orig': y_v})
    es_df['es_res'] = es_df['orig'].ewm(alpha=alpha, adjust=False).mean()
    es_al_a = np.zeros(len(y_v) + 1)
    es_al_a[0] = es_df['es_res'][0]
    es_al_a[1] = es_df['es_res'][0]
    for i in range(2, len(y_v) + 1):
        es_al_a[i] = es_df['es_res'][i-1]
    t1 = np.append(df[time_name].dropna().values[:len(y_v)], [
                   df[time_name].dropna().values[:len(y_v)][-1] + 1])
    org_data = y_v.tolist()
    es_df_e = pd.DataFrame(
        {time_name: t1, y_name: np.append(y_v, [np.nan]), f'ES({alpha})': es_al_a})
    return pd.concat([es_df_e, df.drop(columns=[time_name, y_name])], axis=1)


def exp_trend(df, alpha, beta, average, trend, n=2, y_name="", time_name="", drop=True):
    """
>>> esm_df = mgt2001.ts.exp_trend(esm_df, 0.3, 0.3, 38, 1, n=2, y_name='Enrollments', time_name='Time')
>>> esm_df = esm_df.merge(mgt2001.ts.exp_trend(df, 0.3, 0.5, 38, 1, n=2, y_name='Enrollments', time_name='Time'))

    Set drop to False, if a full trend DataFrame is needed
    """
    from statsmodels.tsa.api import Holt
    import statsmodels.api as sm

    y_v = df[y_name].dropna().values
    esm_a = np.array(y_v)
    esm_model = Holt(esm_a, initialization_method='known', initial_level=average, initial_trend=trend).fit(
        smoothing_level=alpha, smoothing_trend=beta, optimized=False)
    esm_fit = esm_model.fittedvalues
    esm_fcast = esm_model.forecast(n)
    esm_ab_a = np.zeros(len(y_v) + n + 1)
    esm_ab_a[0] = average + trend
    for i in range(1, len(y_v) + 1):
        esm_ab_a[i] = esm_fit[i-1]
    for i in range(len(y_v) + 1, len(y_v) + n + 1):
        esm_ab_a[i] = esm_fcast[i-len(y_v)-1]

    Ini_v = average + trend
    if drop:
        t1 = np.append(df[time_name].dropna().values[:len(y_v)], [
            df[time_name].dropna().values[:len(y_v)][-1] + i for i in range(1, n + 1)])
        org_data = np.append(y_v, [np.nan]*n)
        wsm_df_ab = pd.DataFrame(
            {time_name: t1, f'{y_name}': org_data, f'EST({alpha}, {beta}, {average}, {trend})': esm_ab_a[1:]})
        old_df = df.drop(columns=[time_name, y_name])
        # na_df = pd.DataFrame(
        #     [[np.nan for i in range(len(old_df.columns))]], columns=old_df.columns)
        # old_df = pd.concat([na_df, old_df], ignore_index=True)
        for i in range(n):
            na_df = pd.DataFrame(
                [[np.nan for i in range(len(old_df.columns))]], columns=old_df.columns)
            old_df = pd.concat([old_df, na_df], ignore_index=True)
    else:
        t1 = np.append(df[time_name].dropna().values[:len(y_v)], [
            df[time_name].dropna().values[:len(y_v)][-1] + i for i in range(1, n + 2)])
        W_ini_v = np.append(Ini_v, y_v)
        org_data = np.append(W_ini_v, [np.nan]*n)
        wsm_df_ab = pd.DataFrame(
            {time_name: t1, f'{y_name}': org_data, f'EST({alpha}, {beta}, {average}, {trend})': esm_ab_a[:]})
        old_df = df.drop(columns=[time_name, y_name])
        na_df = pd.DataFrame(
            [[np.nan for i in range(len(old_df.columns))]], columns=old_df.columns)
        old_df = pd.concat([na_df, old_df], ignore_index=True)
        for i in range(n):
            na_df = pd.DataFrame(
                [[np.nan for i in range(len(old_df.columns))]], columns=old_df.columns)
            old_df = pd.concat([old_df, na_df], ignore_index=True)

    return pd.concat([wsm_df_ab, old_df], axis=1)


def error_metric(y=None, pred_y=None, df: pd.DataFrame = None, measurements: list = None, y_name: str = '', precision=4):
    def compute_error(y, pred_y):
        raw_err = np.array(y - pred_y)
        err = raw_err[~np.isnan(raw_err)]
        raw_percentage = np.array((y - pred_y) / y)
        err_percentage = raw_percentage[~np.isnan(raw_percentage)]

        MAD = np.absolute(err).mean()
        MSE = np.mean(err * err)
        RMSE = np.sqrt(np.mean(err * err))
        MAPE = np.absolute(err_percentage).mean() * 100

        err_dict = {'MAD': MAD, 'MSE': MSE, 'RMSE': RMSE, 'MAPE': MAPE}

        return err_dict

    if measurements is None:

        err_dict = compute_error(y, pred_y)

        result = f"""======= Error Metrics =======
MAD = {err_dict['MAD']:.{precision}f}
MSE = {err_dict['MSE']:.{precision}f}
RMSE = {err_dict['RMSE']:.{precision}f}
MAPE = {err_dict['MAPE']:.{precision}f}%
"""
        print(result)
        return err_dict
    else:
        err_df = pd.DataFrame(columns=measurements, index=[
                              'MAD', 'MSE', 'RMSE', 'MAPE'])
        df = df.dropna()
        for measurement in measurements:
            y = df[y_name]
            pred_y = df[measurement]
            err_dict = compute_error(y, pred_y)

            result = f"""### Error Metrics - {measurement}
MAD = {err_dict['MAD']:.{precision}f}
MSE = {err_dict['MSE']:.{precision}f}
RMSE = {err_dict['RMSE']:.{precision}f}
MAPE = {err_dict['MAPE']:.{precision}f}%
"""
            print(result)

            err_df[measurement] = list(err_dict.values())

        min_val = err_df.min(axis=1).values

        def _highlight(val):
            color = 'lightgreen' if val in min_val else 'default'
            return 'background-color: %s' % color

        style_df = err_df.style.applymap(_highlight)

        return err_df, style_df
